/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov [committer] */

#pragma once  // include this header only once per compilation unit


#include "CaConstants.h"
#include "CaMeasurementTime.h"
#include "CaMeasurementXy.h"
#include "CaSimd.h"
#include "CaTrackFit.h"
#include "CaTrackParam.h"
#include "CbmDefs.h"

#include <vector>

class CbmMvdHit;
class CbmStsHit;
class CbmStsTrack;
class CbmGlobalTrack;
class CbmKFVertex;
class CbmMuchPixelHit;
class CbmTrdHit;
class CbmTofHit;

class FairTrackParam;
class TClonesArray;

namespace cbm::algo::ca
{
  class FieldRegion;
}  // namespace cbm::algo::ca

namespace
{
  using namespace cbm::algo;
}

/// A fitter for the Cbm tracks
///
class CbmKfTrackFitter {
 public:
  /// A node on the trajectory where the track parameters are:
  /// a) measured and / or
  /// b) scattered and / or
  /// c) need to be estimated
  /// The nodes must be ordered by increasing Z
  ///
  struct FitNode {

    double fZ{0.};  ///< Z coordinate

    ca::TrackParamV fTrack{};  ///< fitted track

    /// == Material information (if present)
    // TODO: change to the material layer index when the material layer is implemented
    int fMaterialLayer{-1};  ///< index of the material layer. Currently equal to the active tracking station index

    double fRadThick{0.};  ///< material radiation thickness at fZ

    /// == Hit information ( if present )

    ca::MeasurementXy<ca::fvec> fMxy{};  ///< XY-measurement at fZ

    ca::MeasurementTime<ca::fvec> fMt{};  ///< time measurement at fZ

    ECbmModuleId fHitSystemId{ECbmModuleId::kNotExist};  ///< detector system ID of the hit
    int fHitAddress{-1};                                 ///< detector ID of the hit
    int fHitIndex{-1};                                   ///< hit index in the detector hit array

    /// == Flags etc
    bool fIsXySet{false};        ///< true if the XY measurement is set
    bool fIsTimeSet{false};      ///< true if the time measurement is set
    bool fIsFitted{false};       ///< true if the node is fitted, false if the fit failed
    bool fIsRadThickSet{false};  ///< true if the radiation thickness is set
    int fReference1{-1};         ///< some reference can be set by the user
    int fReference2{-1};         ///< some reference can be set by the user
  };

  /// A track to be fitted
  struct Track {
    std::vector<FitNode> fNodes;  ///< nodes on the track
    int fFirstHitNode{-1};        ///< index of the first node with the XY measurement
    int fLastHitNode{-1};         ///< index of the last node with the XY measurement

    /// externally defined inverse momentum for the Multiple Scattering calculation.
    /// It is used for the tracks in field-free regions.
    /// When the momentum can be fitted, the fitted value is used.
    /// the default value is set to 0.1 GeV/c
    double fMsQp0{1. / 0.1};
    bool fIsMsQp0Set{false};

    void MakeConsistent();  // make the structure fields consistent
  };

  CbmKfTrackFitter();
  ~CbmKfTrackFitter();

  /// initialize the fitter. It must be called in the Init() of the user task.
  /// when called later, it sees track branches always empty for whatever reason
  void Init();

  /// set particle hypothesis (mass and electron flag) via particle PDG
  void SetParticleHypothesis(int pid);

  /// set particle  mass
  void SetMassHypothesis(double mass);

  /// set electron flag (bremmstrallung will be applied)
  void SetElectronFlag(bool isElectron);

  bool CreateMvdStsTrack(Track& kfTrack, int stsTrackIndex);
  bool CreateGlobalTrack(Track& kfTrack, int globalTrackIndex);
  bool CreateGlobalTrack(Track& kfTrack, const CbmGlobalTrack& globalTrack);

  /// fit the track
  void FitTrack(CbmKfTrackFitter::Track& t);

  /// fit sts tracks
  // void FitStsTracks(vector<CbmStsTrack>& Tracks, const vector<int>& pidHypo);

 private:
  void FilterFirstMeasurement(const FitNode& n);
  void AddMaterialEffects(const Track& t, FitNode& n, bool upstreamDirection);
  // combine two tracks
  void Smooth(ca::TrackParamV& t1, const ca::TrackParamV& t2);

  std::tuple<ca::fvec, ca::fvec> Smooth1D(ca::fvec x1, ca::fvec Cxx1, ca::fvec x2, ca::fvec Cxx2);
  std::tuple<ca::fvec, ca::fvec, ca::fvec, ca::fvec, ca::fvec> Smooth2D(ca::fvec x1, ca::fvec y1, ca::fvec Cxx1,
                                                                        ca::fvec Cxy1, ca::fvec Cyy1, ca::fvec x2,
                                                                        ca::fvec y2, ca::fvec Cxx2, ca::fvec Cxy2,
                                                                        ca::fvec Cyy2);

 private:
  // input data arrays
  TClonesArray* fInputMvdHits{nullptr};
  TClonesArray* fInputStsHits{nullptr};
  TClonesArray* fInputMuchHits{nullptr};
  TClonesArray* fInputTrdHits{nullptr};
  TClonesArray* fInputTofHits{nullptr};

  TClonesArray* fInputGlobalTracks{nullptr};
  TClonesArray* fInputStsTracks{nullptr};
  TClonesArray* fInputMuchTracks{nullptr};
  TClonesArray* fInputTrdTracks{nullptr};
  TClonesArray* fInputTofTracks{nullptr};

  //

  bool fIsInitialized = {false};  // is the fitter initialized

  ca::TrackFit fFit;  // track fit object

  double fMass{ca::constants::phys::PionMass};  // mass hypothesis for the fit
  bool fIsElectron{false};                      // fit track as an electron (with the bermsstrallung effect)
};
